=head1 NAME

ESP::Parser - event-driven reader for ESP files

=head1 SYNOPSIS

  package MyReader;

  use base 'ESP::Parser';

  sub h_start_major { ... }
  sub h_major_unknown { ... }
  sub h_end_major { ... }
  sub h_start_minor { ... }
  sub h_end_minor { ... }
  sub h_minor_contents { ... }

  package main;

  my $reader=MyReader->new();
  $reader->read_from_fh(\*STDIN);

=head1 DESCRIPTION

This is an event-driven parser for ESP files.  To use, derive your own
class from it, and implement the methods appropriately.

=cut

package ESP::Parser;

use Carp;
use strict;

sub new {
  my $this=shift;
  my $class=ref($this) || $this;
  my $self={};
  bless $self, $class;
  $self->init(@_);
  return $self;
}

sub init {
}

sub read_bytes {
  my $self=shift;
  my $fh=shift;
  my $len=shift;

  my $buf;
  my $count=read($fh, $buf, $len);
  if ($count != $len) {
    confess "Short read of $count bytes, expected $len";
  }
  return $buf;
}

sub read_from_fh {
  my $self=shift;

  my $fh=shift;

  # just in case ...
  binmode $fh;

  $self->start_parse();

  while (not eof $fh) {
    $self->read_major_from_fh($fh);
  }

  $self->end_parse();
}

sub start_parse { }
sub end_parse { }

sub read_major_from_fh {
  my $self=shift;
  my $fh=shift;

  # the file is made up of a series of major sections, each with a
  # 12-byte length.
  my $sect=$self->read_bytes($fh, 4);

  my $len=unpack("L",$self->read_bytes($fh, 4));

  $self->h_start_major($sect,$len);

  my $unknown=$self->read_bytes($fh, 8);
  $self->h_major_unknown($sect, $unknown);

  my $total=0;
  while ($total < $len) {
    $total += $self->read_minor_from_fh($fh, $sect);
  }

  if ($total != $len) {
    die "overlong section $sect: read $total, wanted $len";
  }

  $self->h_end_major($sect, $len);

  return $total;
}

sub h_start_major { }
sub h_major_unknown { }
sub h_end_major { }

sub read_minor_from_fh {
  my $self=shift;
  my $fh=shift;
  my $major=shift;

  # each major section is made up of a series of minor sections, each with a
  # 4-byte length.
  my $sect=$self->read_bytes($fh, 4);

  my $len=unpack("L",$self->read_bytes($fh, 4));

  $self->h_start_minor($major, $sect, $len);

  my $cont=$self->read_bytes($fh, $len);

  $self->h_minor_contents($major, $sect, $cont);

  $self->h_end_minor($major, $sect, $len);

  return $len + 8; # account for header too
}

sub h_start_minor { }
sub h_end_minor { }
sub h_minor_contents { }

1;
