package ESP::ToXML;

use base 'ESP::Parser';

sub start_parse {
}

sub end_parse { }

sub as_hex {
  my $self=shift;
  my $tag=shift;
  my $bytes=shift;

  # shortcut for all-NULL or empty data area
  if ($bytes =~ m!^\0+$! or length($bytes) == 0) {
    return ("<$tag f=b e=0 l=", length($bytes), "/>\n");
  }

  my @rv="<$tag f=b e=h>";
  push @rv, map { sprintf("%02x ", $_) } unpack("C*", $bytes);
  $rv[$#rv] =~ s!\s*$!!;

  # if it contains a good proportion of printing chars, print out the
  # actual contents in a comment, for testing.
  my $prcnt = ((my $tmp=$bytes) =~ tr/[[:print:]]//);
  if ($prcnt / length($bytes) > 0.1) {
    (my $txt=$bytes) =~ s![^[:print:]]!?!g;
    push @rv, " <!-- ", $txt, " -->";
  }

  # allow attrs in the tag, but not in the end tag
  (my $end=$tag) =~ s!\s.*$!!;
  push @rv, "</$end>\n";
  @rv;
}

sub as_int2 {
  my $self=shift;
  my $tag=shift;
  my $bytes=shift;

  my @rv="<$tag f=b e=i2>";
  push @rv, map { "$_ " } unpack("s*", $bytes);
  $rv[$#rv] =~ s!\s*$!!;

  # allow attrs in the tag, but not in the end tag
  (my $end=$tag) =~ s!\s.*$!!;
  push @rv, "</$end>\n";
  @rv;
}

sub as_int4 {
  my $self=shift;
  my $tag=shift;
  my $bytes=shift;

  my @rv="<$tag f=b e=i4>";
  push @rv, map { "$_ " } unpack("l*", $bytes);
  $rv[$#rv] =~ s!\s*$!!;

  # allow attrs in the tag, but not in the end tag
  (my $end=$tag) =~ s!\s.*$!!;
  push @rv, "</$end>\n";
  @rv;
}

sub as_float {
  my $self=shift;
  my $tag=shift;
  my $bytes=shift;

  if ($self->{exactfp}) {
    return $self->as_hex($tag, $bytes);
  }

  my @rv="<$tag f=b e=f>";

  # conversion from binary float to ASCII is not exact!  So at least
  # check for differences in the eventual binary rep.
  #my @strval;
  #for my $val (unpack("f*", $bytes)) {
  #  my $prec=3; # minimum
  #  my $rv;
  #  do {
  #    $rv = sprintf("%1.${prec}f", $val);
  #    $prec++;
  #    warn "Comparing $val and $rv";
  #  } while (abs($val - $rv) >= 1.0e-44);
  #  push @strval, $rv;
  #}
  #push @rv, map { "$_ " } @strval;

  # actually, we just live with the inaccuracy for now.  Hopefully
  # this is enough SF
  push @rv, map { sprintf("%1.3f ", $_) } unpack("f*", $bytes);

  $rv[$#rv] =~ s!\s*$!!;

  # allow attrs in the tag, but not in the end tag
  (my $end=$tag) =~ s!\s.*$!!;
  push @rv, "</$end>\n";
  @rv;
}

sub mangle_string {
  my $self=shift;
  my $origstring=shift;

  # we strip CRLF from strings.  This is necessary because we're
  # reading in binary mode, but our stdout is in text mode (for XML),
  # so we want to use native string formatting in output.
  (my $string=$origstring) =~ s!\r\n!\n!gs;

  # ... but make sure the reverse coding will work, this might not
  # actually be a string!
  (my $newstring=$string) =~ s!\n!\r\n!gs;
  if ($newstring ne $origstring) {
    die "Coding error: non-CRLF \"string\" '$origstring'";
  }

  # encode thte default HTML-style entities
  $string =~ s!\&!\&amp;!g;
  $string =~ s!<!\&lt;!g;
  $string =~ s!>!\&gt;!g;

  $string;
}

# f=s => string (no NULL terminator)
sub as_string {
  my $self=shift;
  my $tag=shift;
  my $string=shift;
  my @rv="<$tag f=s e=t>";

  push @rv, $self->mangle_string($string);

  # allow attrs in the tag, but not in the end tag
  (my $end=$tag) =~ s!\s.*$!!;
  push @rv, "</$end>\n";
  @rv;
}

# f=z => NULL-terminated string
sub as_nullstring {
  my $self=shift;
  my $tag=shift;
  (my $string=shift) =~ s!\0$!!;
  my @rv="<$tag f=z e=t>";

  push @rv, $self->mangle_string($string);

  # allow attrs in the tag, but not in the end tag
  (my $end=$tag) =~ s!\s.*$!!;
  push @rv, "</$end>\n";
  @rv;
}

# f=p => NULL-padded string area, l gives length
sub as_string_padded {
  my $self=shift;
  my $tag=shift;
  my $string=shift;
  my $len=shift || length($string);

  $string=unpack("Z*", $string);

  my @rv="<$tag f=p e=t l=" . $len . ">";

  push @rv, $self->mangle_string($string);

  # allow attrs in the tag, but not in the end tag
  (my $end=$tag) =~ s!\s.*$!!;
  push @rv, "</$end>\n";
  @rv;
}

sub h_start_major {
  my $self=shift;
  my $major=shift;
  $self->output("<", $self->codemap($major), ">\n");
}
sub h_major_unknown {
  my $self=shift;
  my $major=shift;
  my $unknown=shift;
  $self->output("  ", $self->as_hex('unknown nohdr', $unknown));
}
sub h_end_major {
  my $self=shift;
  my $major=shift;
  $self->output("</$major>\n");
}

sub h_start_minor {
  my $self=shift;
  my $major=shift;
  my $minor=shift;
  #$self->output("  <$minor>\n");
}
sub h_end_minor {
  my $self=shift;
  my $major=shift;
  my $minor=shift;
  #$self->output("  </$minor>\n");
}

sub codemap {
  my $self=shift;
  my $sect=shift;

  my %codemap=
    (
     INTV => 'integer',
     FLTV => 'float',
     STRV => 'string',

     MAST => 'master-file',

     HEDR => 'global-header',

     SCPT => 'script',

     SCHD => 'header',
     SCDT => 'compiled',
     SCTX => 'text',

     LTEX => 'texture',

     STAT => 'static',

     LIGH => 'light',

     NPCO => 'inventory',
     NPCS => 'invnt-spell',

     CNDT => 'container-weight',

    );
  if ($codemap{$sect}) {
    return "$codemap{$sect} code=$sect";
  }
  return $sect;
}

sub h_minor_contents {
  my $self=shift;
  my $major=shift;
  my $minor=shift;

  my $call="d_${major}_${minor}";
  my @dat;
  if ($self->can($call)) {
    # we have a specific method for this type
    @dat=$self->$call(@_);
  } else {
    # see if we can guess what it is
    my $dat=shift;
    my $taginfo=$self->codemap($minor);

    my %typemap=
      (
       FLTV => 'as_float',
       # INTV => as_int4, # don't need this one, it's the default, as
                          # they usually contain only one value
       FLAG => 'as_hex',
       CLDT => 'as_int4',
       XSCL => 'as_float',
       YSCL => 'as_float',
       ZSCL => 'as_float',
       CNDT => 'as_float',
       NPCS => 'as_string_padded',
       NPCO => 'd_NPCO',
      );

    eval {
      if ($typemap{"${major}_${minor}"}) {
        my $meth=$typemap{"${major}_${minor}"};
        @dat=$self->$meth($taginfo, $dat);
      } elsif ($typemap{$minor}) {
        my $meth=$typemap{$minor};
        @dat=$self->$meth($taginfo, $dat);
      } elsif (length($dat) == 1) {
        # single-byte values we always code as hex
        @dat=$self->as_hex($taginfo,$dat);
      } elsif (length($dat) == 2) {
        @dat=$self->as_int2($taginfo,$dat);
      } elsif (length($dat) == 4) {
        @dat=$self->as_int4($taginfo, $dat);
      } elsif ($dat =~ m!^[\n\r\s[:print:]]+$!s) {
        # string
        @dat=$self->as_string($taginfo, $dat);
      } elsif ($dat =~ m!^[\n\r\s[:print:]]+\0$!s) {
        # null-terminated string
        @dat=$self->as_nullstring($taginfo, $dat);
      } else {
        # fall back to binary dump
        @dat=$self->as_hex($taginfo, $dat);
      }
    };
    if ($@) {
      # got an error encoding, we'll binary-dump it anyway
      warn "Warning: Error with guessed pretty coding of $major/$minor: $@";
      @dat=$self->as_hex($taginfo, $dat);
    }
  }
  #map { $_ =~ s!\n!\n  !gm; $_ } @dat;
  #$dat[$#dat] =~ s!  $!!;
  $self->output("  ", @dat);
}

# f=b => arbitary binary data
sub d_TES3_HEDR {
  my $self=shift;
  my @info=unpack("fLZ32Z256L", $_[0]);
  return (
          "<HEDR>\n",
          "    ", $self->as_float('version nohdr', pack("f",$info[0])),
          "    ", $self->as_hex('unknown nohdr', pack("L",$info[1])),
          "    ", $self->as_string_padded('author nohdr', $info[2], 32),
          "    ", $self->as_string_padded('description nohdr', $info[3], 256),
          "    ", $self->as_hex('unknown nohdr', pack("L",$info[4])),
          "  </HEDR>\n",
         );
}

sub d_NPCO {
  my $self=shift;
  my $tag=shift;
  my $data=shift;
  my @info=unpack("lZ*", $data);
  (my $end=$tag) =~ s!\s+.*$!!;
  return (
          "<$tag>\n",
          "    ", $self->as_int4('count nohdr', pack('l',$info[0])),
          "    ", $self->as_string_padded('id nohdr', $info[1], length($data)-4),
          "  </$end>\n"
         );
}

sub output {
  my $self=shift;
  print @_;
}

1;
