#! perl -w

=head1 NAME

xmlespc.pl - convert XML files back to .esp

=head1 SYNOPSIS

  perl espxmlc.pl <file.xml> <file.esp>

=head1 DESCRIPTION

Converts XML files produced by espxmlc.pl into a .esp file.  See
espxmlc.pl for more details.

=head1 AUTHOR

Andy Mortimer, andy.mortimer@zetnet.co.uk

=head1 VERSION

  $Id: xmlespc.pl,v 1.2 2003/05/29 17:11:47 andy Exp $

=head1 BUGS

None known in the XML-to-ESP convertor, but see the list in espxmlc.pl
for the known bugs there.

=cut

use strict;

use FindBin;
use lib "$FindBin::Bin";

package WriteESP;

use base 'ESP::XMLParser';

use IO::Handle;

#use HTML::Parser;
#use base 'HTML::Parser';

#sub new {
#  my $this=shift;
#  my $class=ref($this) || $this;
#  my %args=(
#            @_,
#            api_version => 3,
#            strict_names => 1,
#            xml_mode => 1,
#            unbroken_text => 1,
#            case_sensitive => 1,
#            h_start => 'start', 'self, attr, attrseq, text',
#            h_end   => 'end',   'self, tagname, text',
#            h_text  => 'text',  'self, text, is_cdata',
#           );
#  return $this->HTML::Parser::new(%args);
#}

sub start {
  my $self=shift;
  my $tag=shift;
  my $attr=shift;
  #print "start of $tag, attrs: ", (map { "$_ => ".$attr->{$_}.", " } keys %$attr), "\n";

  # to allow pretty XML, we can record the actual section name in an
  # attribute
  my $sectname=$tag;
  if (defined $attr->{code}) {
    $sectname=$attr->{code};
  }

  if ($attr->{nohdr}) {
    $self->datbegin_nohdr($sectname);
  } else {
    $self->datbegin($sectname);
  }

  # are we reading data?  If so, we'll have an f= attribute
  if (defined $attr->{f}) {
    $self->dat_set_format($attr);
  } else {
    $self->dat_set_format(undef);
  }

  # special case: some formats generate data inline
  if (defined $attr->{e} and $attr->{e} eq '0') {
    $self->rawdata(chr(0) x $attr->{l});
  }
}
sub end {
  my $self=shift;
  my $tag=shift;
  $self->datend($tag);
  # we always stop reading data at the end.  We can do this, as we
  # assume only the innermost objects contain data.  If this changes,
  # the data format needs to be stacked too
  $self->dat_set_format(undef);
}
sub text {
  my $self=shift;
  my $text=shift;
  $self->formdata($text);
}

# stack-based writing, we have to know lengths before we can write
sub datbegin {
  my $self=shift;
  my $sect=shift;
  $self->rawdata($sect);
  unshift @{$self->{datastack}}, "";
  push @{$self->{tagstack}}, 1;
  #print ":: begin $sect\n";
}
sub datbegin_nohdr {
  my $self=shift;
  my $sect=shift;
  # the tags marked 'nohdr' are just for bookkeeping, we ignore their
  # names etc and just take the data
  push @{$self->{tagstack}}, 0;
  #print ":: begin $sect nohdr\n";
}
sub datend {
  my $self=shift;

  #my $sect=shift;
  #print ":: end $sect\n";

  # if we didn't write a header when we started, we're just writing
  # data, so this is all ignored
  my $endsect=pop @{$self->{tagstack}};
  if ($endsect) {
    my $data=shift @{$self->{datastack}};
    # *** evil, evil hack -- the top-level headers don't include the
    # following 8 bytes of "junk" in the count, so if we're at the top
    # level, we will subtract 8 from the length.
    if (@{$self->{datastack}}) {
      $self->rawdata(pack("L",length($data)));
    } else {
      $self->rawdata(pack("L",length($data) - 8));
    }
    $self->rawdata($data);
  }
}

sub dat_set_format {
  my $self=shift;
  $self->{format}=shift;
}

# this call reads formatted data and makes raw data
sub formdata {
  my $self=shift;
  my $dat=shift;

  # skip if we have no format -- that means we're ignoring data
  return unless defined $self->{format};

  my $fmt=$self->{format}{f};
  my $enc=$self->{format}{e};
  if ($fmt eq 'b') {
    # binary data.  All (current!) encodings of binary data are
    # ws-separated terms, so split now.
    my @data=split(/\s+/,$dat);
    if ($enc eq 'f') {
      # floating-point
      $self->rawdata(pack("f*", @data));
    } elsif ($enc eq 'h') {
      # hex bytes
      #use Data::Dumper;
      #print Dumper(\@data, [map { hex } @data], [pack("C*", map { hex } @data)]);
      $self->rawdata(pack("C*", map { hex } @data));
    } elsif ($enc eq 'i2') {
      # short int
      $self->rawdata(pack("s*",@data));
    } elsif ($enc eq 'i4') {
      # short int
      $self->rawdata(pack("l*",@data));
    } else {
      die "unrecognised binary encoding '$enc'";
    }
  } elsif ($fmt eq 's') {
    # string, no NULL terminator => just the bytes
    die "bad string encoding '$enc'" unless $enc eq 't';
    $dat =~ s!\n!\r\n!gs;
    $self->rawdata($dat);
  } elsif ($fmt eq 'z') {
    # string, NULL terminated
    die "bad string encoding '$enc'" unless $enc eq 't';
    $dat =~ s!\n!\r\n!gs;
    $self->rawdata($dat . chr(0));
  } elsif ($fmt eq 'p') {
    # NULL-padded string
    die "bad string encoding '$enc'" unless $enc eq 't';
    my $len=$self->{format}{l};
    die "no length for NULL-padded string" unless defined $len;
    die "string \"$dat\" too long for slot of length $len"
      if length($dat) > $len;
    $dat =~ s!\n!\r\n!gs;
    $self->rawdata(pack("a$len", $dat));
  } else {
    die "unrecognised format '$fmt'";
  }
}

sub rawdata {
  my $self=shift;
  my $data=shift;

  if (defined $self->{datastack} and @{$self->{datastack}}) {
    $self->{datastack}[0] .= $data;
  } else {
    # nothing on the stack, we're at the top level, so print it
    my $outfh=$self->{outfh};
    print  $outfh $data;
    $outfh->flush();
  }
}

package main;

my $parser=new WriteESP();

die "usage: $0 <file.xml> <file.esp>" unless @ARGV == 2;

my $fname=shift @ARGV;
open FILE, "<$fname"
  or die "error reading $fname: $!";

warn "Reading $fname\n";

my $outf=shift @ARGV;
open OUT, ">$outf"
  or die "error writing $outf: $!";

warn "Writing $outf\n";

$parser->{outfh} = \*OUT;
binmode OUT;

$parser->parse_fh(\*FILE);

close FILE;
close OUT 
  or die "error finishing write to $outf: $!";
